/**
 * Abstraction for threading to allow porting.
 * Currently using either pthread or C++ 11 threads.
 */

#ifndef PLS_THREAD_H
#define PLS_THREAD_H

#include <functional>
#include <pthread.h>

namespace pls {
    namespace internal {
        namespace base {
            using thread_entrypoint = void();

            class this_thread {
                template<typename Function, typename State>
                friend class thread;
                static pthread_key_t local_storage_key_;
                static bool local_storage_key_initialized_;

            public:
                static void yield() {
                    pthread_yield();
                }

                template<typename T>
                static T* state() {
                    return reinterpret_cast<T*>(pthread_getspecific(local_storage_key_));
                }

                template<typename T>
                static void set_state(const T& state) {
                    *reinterpret_cast<T*>(pthread_getspecific(local_storage_key_)) = state;
                }
            };

            template<typename Function, typename State>
            class thread {
                friend class this_thread;
                // Keep a copy of the function (lambda) in this object to make sure it is valid when called!
                Function function_;
                // Keep the local state we hold in here
                State state_;

                // Keep handle to native implementation
                pthread_t pthread_thread_;

                static void* start_pthread_internal(void* thread_pointer) {
                    auto my_thread = reinterpret_cast<thread*>(thread_pointer);
                    pthread_setspecific(this_thread::local_storage_key_, (void*)&my_thread->state_);
                    my_thread->function_();
                    pthread_exit(nullptr);
                }

            public:
                explicit thread(const Function& function, const State& state):
                        function_{function},
                        state_{state},
                        pthread_thread_{} {}

                void start() {
                    if (!this_thread::local_storage_key_initialized_) {
                        pthread_key_create(&this_thread::local_storage_key_, nullptr);
                        this_thread::local_storage_key_initialized_ = true;
                    }

                    pthread_create(&pthread_thread_, nullptr, start_pthread_internal, (void *)(this));
                }
            public:
                void join() {
                    pthread_join(pthread_thread_, nullptr);
                }

                // make object move only
                thread(thread&&) noexcept = default;
                thread& operator=(thread&&) noexcept = default;

                thread(const thread&) = delete;
                thread& operator=(const thread&) = delete;
            };

            template<typename Function, typename State>
            thread<Function, State> create_thread(const Function& function, const State& state) {
                return thread<Function, State>(function, state);
            }

        }
    }

}

#endif //PLS_THREAD_H
